"""Module for searchcontext for collection of cases."""

from fmu.sumo.explorer.objects._search_context import SearchContext


class Cases(SearchContext):
    def __init__(self, sc, uuids):
        super().__init__(
            sc._sumo, must=[{"terms": {"fmu.case.uuid.keyword": uuids}}]
        )
        self._hits = uuids
        return

    def _maybe_prefetch(self, index):
        return

    async def _maybe_prefetch_async(self, index):
        return

    def filter(self, **kwargs):
        sc = super().filter(**kwargs)
        uuids = sc.uuids
        return Cases(self, uuids)



"""Module containing class for cpgrid"""

from typing import Dict

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._child import Child
from fmu.sumo.explorer.objects._search_context import SearchContext


class CPGrid(Child):
    """Class representing a cpgrid object in Sumo."""

    def __init__(self, sumo: SumoClient, metadata: Dict, blob=None) -> None:
        """
        Args:
            sumo (SumoClient): connection to Sumo
            metadata (dict): dictionary metadata
            blob: data object
        """
        super().__init__(sumo, metadata, blob)

    def to_cpgrid(self):
        """Get cpgrid object as a Grid
        Returns:
            Grid: A Grid object
        """
        try:
            from xtgeo import grid_from_file
        except ModuleNotFoundError:
            raise RuntimeError(
                "Unable to import xtgeo; probably not installed."
            )
        try:
            return grid_from_file(self.blob)
        except TypeError as type_err:
            raise TypeError(f"Unknown format: {self.format}") from type_err

    async def to_cpgrid_async(self):
        """Get cpgrid object as a Grid
        Returns:
            Grid: A Grid object
        """
        try:
            from xtgeo import grid_from_file
        except ModuleNotFoundError:
            raise RuntimeError(
                "Unable to import xtgeo; probably not installed."
            )

        try:
            return grid_from_file(await self.blob_async)
        except TypeError as type_err:
            raise TypeError(f"Unknown format: {self.format}") from type_err

    @property
    def grid_properties(self):
        """Get cpgrid_property instances that use this cpgrid instance.
        Returns:
            GridProperties: a SearchContext that holds the linked CPGridProperty instances.
        """
        sc = SearchContext(self._sumo)
        return sc.filter(
            complex={
                "bool": {
                    "must": [
                        {
                            "term": {
                                "data.geometry.relative_path.keyword": self._metadata[
                                    "file"
                                ]["relative_path"]
                            }
                        },
                        {
                            "term": {
                                "fmu.case.uuid.keyword": self._metadata["fmu"][
                                    "case"
                                ]["uuid"]
                            }
                        },
                    ]
                }
            }
        )



"""module containing class for table"""

from fmu.sumo.explorer._utils import Utils
from fmu.sumo.explorer.objects.case import Case
from fmu.sumo.explorer.objects.table import Table


class AggregatedTable:
    """Class for representing an aggregated table in Sumo"""

    def __init__(
        self,
        case: Case,
        name: str,
        tag: str,
        iteration: str,
        aggregation: str = "collection",
    ) -> None:
        """Init of aggregated table

        Args:
            case (Sumo.Case): given case object
            name (str): name of table
            tag (str): name of tag
            iteration (str): name of interation
            aggregation (str, optional): aggregation type
        """
        self._sumo = case._sumo
        self._utils = Utils(self._sumo)
        self._case = case
        self._name = name
        self._tag = tag
        self._iteration = iteration
        self._aggregation = aggregation
        self._parameters = None
        self._collection = case.tables.filter(
            name=name,
            tagname=tag,
            iteration=iteration,
            aggregation=aggregation,
        )

    @property
    def columns(self):
        """Return column names

        Returns:
            list: the column names available
        """
        return self._collection.columns

    @property
    def parameters(self):
        """Return parameter set for iteration

        Returns:
            dict: parameters connected to iteration
        """
        if not self._parameters:
            must = self._utils.build_terms(
                {
                    "class.keyword": "table",
                    "_sumo.parent_object.keyword": self._case.uuid,
                    "data.name.keyword": self._name,
                    "data.tagname.keyword": self._tag,
                    "fmu.iteration.name.keyword": self._iteration,
                    "fmu.aggregation.operation.keyword": "collection",
                }
            )

            query = {
                "size": 1,
                "_source": ["fmu.iteration.parameters"],
                "query": {"bool": {"must": must}},
            }

            res = self._sumo.post("/search", json=query)
            doc = res.json()["hits"]["hits"][0]
            self._parameters = doc["_source"]["fmu"]["iteration"]["parameters"]

        return self._parameters

    @property
    async def parameters_async(self):
        """Return parameter set for iteration

        Returns:
            dict: parameters connected to iteration
        """
        if not self._parameters:
            must = self._utils.build_terms(
                {
                    "class.keyword": "table",
                    "_sumo.parent_object.keyword": self._case.uuid,
                    "data.name.keyword": self._name,
                    "data.tagname.keyword": self._tag,
                    "fmu.iteration.name.keyword": self._iteration,
                    "fmu.aggregation.operation.keyword": "collection",
                }
            )

            query = {
                "size": 1,
                "_source": ["fmu.iteration.parameters"],
                "query": {"bool": {"must": must}},
            }

            res = await self._sumo.post_async("/search", json=query)
            doc = res.json()["hits"]["hits"][0]
            self._parameters = doc["_source"]["fmu"]["iteration"]["parameters"]

        return self._parameters

    def __len__(self):
        return len(self._collection)

    def __getitem__(self, column) -> Table:
        try:
            return self._collection.filter(column=column)[0]
        except IndexError as i_ex:
            raise IndexError(
                f"Column: '{column}' does not exist, try again"
            ) from i_ex



"""Contains class for one document"""

import re
from typing import Dict, List

_path_split_rx = re.compile(r"\]\.|\.|\[")


def _splitpath(path):
    parts = _path_split_rx.split(path)
    return [int(x) if re.match(r"\d+", x) else x for x in parts]


def _makeprop(attribute):
    path = _splitpath(attribute)
    return lambda self: self._get_property(path)


class Document:
    """Class for representing a document in Sumo"""

    def __init__(self, metadata: Dict) -> None:
        self._uuid = metadata["_id"]
        self._metadata = metadata["_source"]

    @property
    def uuid(self):
        """Return uuid

        Returns:
            str: the uuid of the case
        """
        return self._uuid

    @property
    def metadata(self):
        """Return metadata for document

        Returns:
            dict: the metadata
        """
        return self._metadata

    def _get_property(self, path: List[str]):
        curr = self._metadata

        for key in path:
            if (isinstance(curr, list) and key < len(curr)) or key in curr:
                curr = curr[key]
            else:
                return None

        return curr

    def __getitem__(self, key: str):
        return self._metadata[key]

    @staticmethod
    def map_properties(cls, propmap):
        for name, attribute, doc in propmap:
            setattr(cls, name, property(_makeprop(attribute), None, None, doc))



import os


def pytest_addoption(parser):
    parser.addoption("--token", action="store", default="")


def pytest_generate_tests(metafunc):
    # token = metafunc.config.option.token
    token = os.environ.get("ACCESS_TOKEN")
    token = token if token and len(token) > 0 else None

    if "token" in metafunc.fixturenames:
        metafunc.parametrize("token", [token])



"""Module containing class for cube object"""

from typing import Dict

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._child import Child


class Cube(Child):
    """Class representig a seismic cube object in Sumo"""

    def __init__(self, sumo: SumoClient, metadata: Dict, blob=None) -> None:
        """
        Args:
            sumo (SumoClient): connection to Sumo
            metadata (dict): cube metadata
        """
        super().__init__(sumo, metadata, blob)
        self._url = None
        self._sas = None

    def _populate_url(self):
        res = self._sumo.get(f"/objects('{self.uuid}')/blob/authuri")
        try:
            res = res.json()
            self._url = res.get("baseuri") + self.uuid
            self._sas = res.get("auth")
        except Exception:
            self._url = res.text

    async def _populate_url_async(self):
        res = await self._sumo.get_async(
            f"/objects('{self.uuid}')/blob/authuri"
        )
        try:
            res = res.json()
            self._url = res.get("baseuri") + self.uuid
            self._sas = res.get("auth")
        except Exception:
            self._url = res.text

    @property
    def url(self) -> str:
        if self._url is None:
            self._populate_url()
        if self._sas is None:
            return self._url
        else:
            return self._url.split("?")[0] + "/"

    @property
    async def url_async(self) -> str:
        if self._url is None:
            await self._populate_url_async()
        if self._sas is None:
            return self._url
        else:
            return self._url.split("?")[0] + "/"

    @property
    def sas(self) -> str:
        if self._url is None:
            self._populate_url()
        if self._sas is None:
            return self._url.split("?")[1]
        else:
            return self._sas

    @property
    async def sas_async(self) -> str:
        if self._url is None:
            await self._populate_url_async()
        if self._sas is None:
            return self._url.split("?")[1]
        else:
            return self._sas

    @property
    def openvds_handle(self):
        try:
            import openvds
        except ModuleNotFoundError:
            raise RuntimeError(
                "Unable to import openvds; probably not installed."
            )

        if self._url is None:
            self._populate_url()

        if self._sas is None:
            return openvds.open(self._url)
        else:
            url = "azureSAS" + self._url[5:] + "/"
            sas = "Suffix=?" + self._sas
            return openvds.open(url, sas)

    @property
    async def openvds_handle_async(self):
        try:
            import openvds
        except ModuleNotFoundError:
            raise RuntimeError(
                "Unable to import openvds; probably not installed."
            )

        if self._url is None:
            await self._populate_url_async()

        if self._sas is None:
            return openvds.open(self._url)
        else:
            url = "azureSAS" + self._url[5:] + "/"
            sas = "Suffix=?" + self._sas
            return openvds.open(url, sas)



"""Module for (pseudo) realization class."""

from typing import Dict

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._document import Document
from fmu.sumo.explorer.objects._search_context import SearchContext

_prop_desc = [
    ("iterationname", "fmu.iteration.name", "FMU iteration name"),
    ("iterationuuid", "fmu.iteration.uuid", "FMU iteration uuid"),
    ("casename", "fmu.case.name", "FMU case name"),
    ("caseuuid", "fmu.case.uuid", "FMU case uuid"),
    ("user", "fmu.case.user.id", "Name of user who uploaded iteration."),
    ("asset", "access.asset.name", "Case asset"),
    ("field", "masterdata.smda.field[0].identifier", "Case field"),
]


class Realization(Document, SearchContext):
    """Class for representing a realization in Sumo."""

    def __init__(self, sumo: SumoClient, metadata: Dict):
        Document.__init__(self, metadata)
        SearchContext.__init__(
            self,
            sumo,
            must=[{"term": {"fmu.realization.uuid.keyword": self.uuid}}],
        )


Realization.map_properties(Realization, _prop_desc)



from collections import deque
from threading import Lock

# Based on https://llego.dev/posts/implement-lru-cache-python/


class LRUCache:
    def __init__(self, capacity):
        self.cache = {}
        self.capacity = capacity
        self.access = deque()
        self.lock = Lock()

    def get(self, key):
        with self.lock:
            if key not in self.cache:
                return None
            else:
                self.access.remove(key)
                self.access.append(key)
                return self.cache[key]

    def put(self, key, value):
        with self.lock:
            if key in self.cache:
                self.access.remove(key)
            elif len(self.cache) == self.capacity:
                oldest = self.access.popleft()
                del self.cache[oldest]
        self.cache[key] = value
        self.access.append(key)

    def has(self, key):
        return key in self.cache

    def clear(self):
        with self.lock:
            self.cache.clear()
            self.access.clear()



"""module containing class for table"""

import logging
from typing import Dict

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._child import Child


class Table(Child):
    """Class representing a table object in Sumo"""

    def __init__(self, sumo: SumoClient, metadata: Dict, blob=None) -> None:
        """
        Args:
            sumo (SumoClient): connection to Sumo
            metadata: (dict): child object metadata
        """
        super().__init__(sumo, metadata, blob)
        self._dataframe = None
        self._arrowtable = None
        self._logger = logging.getLogger("__name__" + ".Table")

    def _get_blob(self):
        if self._blob is None:
            self._blob = self.blob
        return self._blob

    async def _get_blob_async(self):
        if self._blob is None:
            self._blob = await self.blob_async
        return self._blob

    def _read_table(self):
        return self._construct_table_from_blob(self._get_blob())

    async def _read_table_async(self):
        return self._construct_table_from_blob(await self._get_blob_async())

    def _construct_table_from_blob(self, blob):
        import pandas as pd
        import pyarrow.feather as pf

        try:
            if self.dataformat == "csv":
                dataframe = pd.read_csv(blob)
            elif self.dataformat == "parquet":
                dataframe = pd.read_parquet(blob)
            elif self.dataformat == "arrow":
                dataframe = pf.read_feather(blob)
            else:
                raise TypeError(
                    f"Don't know how to convert a blob of format {self.dataformat} to a pandas table."
                )
        except Exception:
            try:
                dataframe = pd.read_csv(blob)
            except Exception:
                try:
                    dataframe = pd.read_parquet(blob)
                except Exception:
                    try:
                        dataframe = pf.read_feather(blob)
                    except Exception:
                        raise TypeError(
                            f"Unable to convert a blob of format {self.dataformat} to pandas table; tried csv, parquet and feather."
                        )
                    pass
                pass
            pass
        return dataframe

    def to_pandas(self):
        """Return object as a pandas DataFrame

        Returns:
            DataFrame: A DataFrame object
        """
        if self._dataframe is None:
            self._dataframe = self._read_table()
        return self._dataframe

    async def to_pandas_async(self):
        """Return object as a pandas DataFrame

        Returns:
            DataFrame: A DataFrame object
        """
        if self._dataframe is None:
            self._dataframe = await self._read_table_async()
        return self._dataframe

    def _read_arrow(self):
        return self._construct_arrow_from_blob(self._get_blob())

    async def _read_arrow_async(self):
        return self._construct_arrow_from_blob(await self._get_blob_async())

    def _construct_arrow_from_blob(self, blob):
        import pandas as pd
        import pyarrow as pa
        import pyarrow.feather as pf
        import pyarrow.parquet as pq

        try:
            if self.dataformat == "csv":
                arrowtable = pa.Table.from_pandas(pd.read_csv(blob))
            elif self.dataformat == "parquet":
                arrowtable = pq.read_table(blob)
            elif self.dataformat == "arrow":
                arrowtable = pf.read_table(blob)
            else:
                raise TypeError(
                    f"Don't know how to convert a blob of format {self.dataformat} to a pandas table."
                )
        except Exception:
            try:
                arrowtable = pa.Table.from_pandas(pd.read_csv(blob))
            except Exception:
                try:
                    arrowtable = pq.read_table(blob)
                except Exception:
                    try:
                        arrowtable = pf.read_table(blob)
                    except Exception:
                        raise TypeError(
                            f"Unable to convert a blob of format {self.dataformat} to arrow; tried csv, parquet and feather."
                        )
                    pass
                pass
            pass
        return arrowtable

    def to_arrow(self):
        """Return object as an arrow Table

        Returns:
            pa.Table: _description_
        """
        if self._arrowtable is None:
            self._arrowtable = self._read_arrow()
        return self._arrowtable

    async def to_arrow_async(self):
        """Return object as an arrow Table

        Returns:
            pa.Table: _description_
        """
        if self._arrowtable is None:
            self._arrowtable = await self._read_arrow_async()
        return self._arrowtable



"""Tests explorer"""

import sys

if not sys.platform.startswith("darwin") and sys.version_info < (3, 12):
    import openvds
import json
import logging
from pathlib import Path
from uuid import UUID

import pytest
from context import (
    Case,
    Explorer,
    SearchContext,
)
from sumo.wrapper import SumoClient
from xtgeo import RegularSurface

TEST_DATA = Path("data")
logging.basicConfig(level="DEBUG")
LOGGER = logging.getLogger()


@pytest.fixture(name="the_logger")
def fixture_the_logger():
    """Defining a logger"""
    return LOGGER  # ut.init_logging("tests", "debug")


@pytest.fixture(name="case_name")
def fixture_case_name() -> str:
    """Returns case name"""
    return "drogon_design_small-2023-01-18"


@pytest.fixture(name="case_uuid")
def fixture_case_uuid() -> str:
    """Returns case uuid"""
    return "2c2f47cf-c7ab-4112-87f9-b4797ec51cb6"


@pytest.fixture(name="seismic_case_uuid")
def fixture_seismic_case_uuid() -> str:
    """Returns seismic case uuid"""
    return "c616019d-d344-4094-b2ee-dd4d6d336217"


@pytest.fixture(name="explorer")
def fixture_explorer(token: str) -> Explorer:
    """Returns explorer"""
    return Explorer("dev", token=token)


@pytest.fixture(name="test_case")
def fixture_test_case(explorer: Explorer, case_name: str) -> Case:
    """Basis for test of method get_case_by_name for Explorer,
    but also other attributes
    """
    return explorer.cases.filter(name=case_name)[0]


@pytest.fixture(name="sumo_client")
def fixture_sumo_client(token: str):
    """Returns SumoClient for dev env"""
    return SumoClient("dev", token=token)


def write_json(result_file, results):
    """writes json files to disc
    args:
    result_file (str): path to file relative to TEST_DATA
    """
    result_file = TEST_DATA / result_file
    with open(result_file, "w", encoding="utf-8") as json_file:
        json.dump(results, json_file)


def read_json(input_file):
    """read json from disc
    args:
    result_file (str): path to file relative to TEST_DATA
    returns:
    content (dict): results from file
    """
    result_file = TEST_DATA / input_file
    with open(result_file, "r", encoding="utf-8") as json_file:
        contents = json.load(json_file)
    return contents


def assert_correct_uuid(uuid_to_check, version=4):
    """Checks if uuid has correct structure
    args:
    uuid_to_check (str): to be checked
    version (int): what version of uuid to compare to
    """
    # Concepts stolen from stackoverflow.com
    # questions/19989481/how-to-determine-if-a-string-is-a-valid-v4-uuid
    type_mess = f"{uuid_to_check} is not str ({type(uuid_to_check)}"
    assert isinstance(uuid_to_check, str), type_mess
    works_for_me = True
    try:
        UUID(uuid_to_check, version=version)
    except ValueError:
        works_for_me = False
    structure_mess = f"{uuid_to_check}, does not have correct structure"
    assert works_for_me, structure_mess


def assert_uuid_dict(uuid_dict):
    """Tests that dict has string keys, and valid uuid's as value
    args:
    uuid_dict (dict): dict to test
    """
    for key in uuid_dict:
        assert_mess = f"{key} is not of type str"
        assert isinstance(key, str), assert_mess
        assert_correct_uuid(uuid_dict[key])


def assert_dict_equality(results, correct):
    """Asserts whether two dictionaries are the same
    args:
    results (dict): the one to check
    correct (dict): the one to compare to
    """
    incorrect_mess = (
        f"the dictionary produced ({results}) is not equal to \n"
        + f" ({correct})"
    )
    assert results == correct, incorrect_mess


def test_get_cases(explorer: Explorer):
    """Test the get_cases method."""

    cases = explorer.cases
    assert isinstance(cases, SearchContext)
    assert isinstance(cases[0], Case)


def test_get_cases_fields(explorer: Explorer):
    """Test SearchContext.filter method with the field argument.

    Shall be case insensitive.
    """

    cases = explorer.cases.filter(field="DROGON")
    for case in cases:
        assert case.field.lower() == "drogon"


def test_get_cases_status(explorer: Explorer):
    """Test the SearchContext.filter method with the status argument."""

    cases = explorer.cases.filter(status="keep")
    for case in cases:
        assert case.status == "keep"


def test_get_cases_user(explorer: Explorer):
    """Test the SearchContext.filter method with the user argument."""

    cases = explorer.cases.filter(user="peesv")
    for case in cases:
        assert case.user == "peesv"


def test_get_cases_combinations(explorer: Explorer):
    """Test the SearchContext.filter method with combined arguments."""

    cases = explorer.cases.filter(
        field=["DROGON", "JOHAN SVERDRUP"],
        user=["peesv", "dbs"],
        status=["keep"],
    )
    for case in cases:
        assert (
            case.user in ["peesv", "dbs"]
            and case.field in ["DROGON", "JOHAN SVERDRUP"]
            and case.status == "keep"
        )


def test_case_surfaces_type(test_case: Case):
    """Test that all objects in Case.surfaces class surface"""
    # assert all([x.metadata["class"] == "surface" for x in test_case.surfaces])
    classes = test_case.surfaces._get_field_values("class.keyword")
    assert len(classes) == 1
    assert classes[0] == "surface"


def test_case_surfaces_size(test_case: Case):
    """Test that Case.surfaces has the correct size"""
    assert len(test_case.surfaces) == 271


def test_case_surfaces_filter(test_case: Case):
    """Test that Case.surfaces has the correct size"""
    # filter on iteration stage
    agg_surfs = test_case.surfaces.filter(stage="iteration")
    assert len(agg_surfs) == 59

    agg_surfs = test_case.surfaces.filter(aggregation=True)
    assert len(agg_surfs)

    # filter on realization stage
    real_surfs = test_case.surfaces.filter(stage="realization")
    assert len(real_surfs) == 212

    real_surfs = test_case.surfaces.filter(realization=True)
    assert len(real_surfs) == 212

    # filter on iteration
    real_surfs = real_surfs.filter(iteration="iter-0")
    assert len(real_surfs) == 212

    # for surf in real_surfs:
    #     assert surf.iteration == "iter-0"
    its = real_surfs._get_field_values("fmu.iteration.name.keyword")
    assert len(its) == 1 and its[0] == "iter-0"

    # filter on name
    non_valid_name_surfs = real_surfs.filter(name="___not_valid")
    assert len(non_valid_name_surfs) == 0

    real_surfs = real_surfs.filter(name="Valysar Fm.")
    assert len(real_surfs) == 56

    # for surf in real_surfs:
    #     assert surf.iteration == "iter-0"
    #     assert surf.name == "Valysar Fm."
    its = real_surfs._get_field_values("fmu.iteration.name.keyword")
    assert len(its) == 1 and its[0] == "iter-0"
    names = real_surfs._get_field_values("data.name.keyword")
    assert len(names) == 1 and names[0] == "Valysar Fm."

    # filter on content
    non_valid_content_surfs = real_surfs.filter(content="___not_valid")
    assert len(non_valid_content_surfs) == 0

    real_surfs = real_surfs.filter(content="depth")
    assert len(real_surfs) == 56

    # filter on tagname
    non_valid_tagname_surfs = real_surfs.filter(tagname="___not_valid")
    assert len(non_valid_tagname_surfs) == 0

    real_surfs = real_surfs.filter(tagname="FACIES_Fraction_Channel")
    assert len(real_surfs) == 4

    # for surf in real_surfs:
    #     assert surf.iteration == "iter-0"
    #     assert surf.name == "Valysar Fm."
    #     assert surf.tagname == "FACIES_Fraction_Channel"
    its = real_surfs._get_field_values("fmu.iteration.name.keyword")
    assert len(its) == 1 and its[0] == "iter-0"
    names = real_surfs._get_field_values("data.name.keyword")
    assert len(names) == 1 and names[0] == "Valysar Fm."
    tagnames = real_surfs._get_field_values("data.tagname.keyword")
    assert len(tagnames) == 1 and tagnames[0] == "FACIES_Fraction_Channel"

    # filter on data format
    non_valid_format_surfs = real_surfs.filter(dataformat="___not_valid")
    assert len(non_valid_format_surfs) == 0

    real_surfs = real_surfs.filter(dataformat="irap_binary")
    assert len(real_surfs) == 4

    # filter on realization
    real_surfs = real_surfs.filter(realization=0)
    assert len(real_surfs) == 1

    assert real_surfs[0].iteration == "iter-0"
    assert real_surfs[0].name == "Valysar Fm."
    assert real_surfs[0].tagname == "FACIES_Fraction_Channel"
    assert real_surfs[0].realization == 0
    assert isinstance(real_surfs[0].to_regular_surface(), RegularSurface)


def test_case_surfaces_pagination(test_case: Case):
    """Test the pagination logic of SurfaceCollection (DocumentCollection)"""
    surfs = test_case.surfaces
    count = 0

    for _ in surfs:
        count += 1

    assert count == len(surfs)


def test_get_case_by_uuid(explorer: Explorer, case_uuid: str, case_name: str):
    """Test that explorer.get_case_by_uuid returns the specified case"""
    case = explorer.get_case_by_uuid(case_uuid)

    assert isinstance(case, Case)
    assert case.uuid == case_uuid
    assert case.name == case_name


@pytest.mark.skipif(
    sys.platform.startswith("darwin") or sys.version_info > (3, 11),
    reason="do not run OpenVDS SEGYImport on mac os or python 3.12",
)
def test_seismic_case_by_uuid(explorer: Explorer, seismic_case_uuid: str):
    """Test that explorer returns openvds compatible cubes for seismic case"""
    case = explorer.get_case_by_uuid(seismic_case_uuid)

    assert isinstance(case, Case)
    assert case.uuid == seismic_case_uuid
    assert len(case.cubes) == 10
    cube = case.cubes[0]
    openvds_handle = cube.openvds_handle

    layout = openvds.getLayout(openvds_handle)  # type: ignore
    channel_count = layout.getChannelCount()
    assert channel_count == 3
    channel_list = []
    for i in range(channel_count):
        channel_list.append(layout.getChannelName(i))
    assert "Amplitude" in channel_list
    assert "Trace" in channel_list
    assert "SEGYTraceHeader" in channel_list


def test_grids_and_properties(explorer: Explorer):
    cases_with_grids = explorer.grids.cases.filter(status="keep")
    cases_with_gridprops = explorer.grid_properties.cases.filter(status="keep")
    cgs = {case.uuid for case in cases_with_grids}
    cgps = {case.uuid for case in cases_with_gridprops}
    assert cgs == cgps
    case = cases_with_grids[0]
    grids = case.grids
    gridprops = case.grid_properties
    xtgrid = grids[0].to_cpgrid()
    gridspec = grids[0].metadata["data"]["spec"]
    assert xtgrid.nlay == gridspec["nlay"]
    assert xtgrid.nrow == gridspec["nrow"]
    assert xtgrid.ncol == gridspec["ncol"]
    xtgridprop = gridprops[0].to_cpgrid_property()
    gridpropspec = gridprops[0].metadata["data"]["spec"]
    assert xtgridprop.nlay == gridpropspec["nlay"]
    assert xtgridprop.nrow == gridpropspec["nrow"]
    assert xtgridprop.ncol == gridpropspec["ncol"]


def test_search_context_select(test_case: Case):
    surfs = test_case.surfaces.filter(realization=True)
    assert "_sumo" in surfs[0].metadata
    surfs.select("fmu")
    assert "_sumo" not in surfs[0].metadata
    assert "fmu" in surfs[0].metadata
    surfs.select(["fmu"])
    assert "_sumo" not in surfs[0].metadata
    assert "fmu" in surfs[0].metadata
    surfs.select({"excludes": ["fmu"]})
    assert "_sumo" in surfs[0].metadata
    assert "fmu" not in surfs[0].metadata
    surfs.select({"includes": ["_sumo"], "excludes": ["_sumo.timestamp"]})
    assert "_sumo" in surfs[0].metadata
    assert "fmu" not in surfs[0].metadata
    assert "timestamp" not in surfs[0].metadata["_sumo"]



import uuid
import warnings
from datetime import datetime
from io import BytesIO
from typing import Dict, List, Tuple

import deprecation
import httpx
from sumo.wrapper import SumoClient

import fmu.sumo.explorer.objects as objects
from fmu.sumo.explorer.cache import LRUCache


def _gen_filter_none():
    def _fn(value):
        return None, None

    return _fn


def _gen_filter_id():
    """Match against document id(s) (in uuid format)."""

    def _fn(value):
        if value is None:
            return None, None
        else:
            return {
                "ids": {
                    "values": value if isinstance(value, list) else [value]
                }
            }, None

    return _fn


def _gen_filter_gen(attr):
    """Match property against either single value or list of values.
    If the value given is a boolean, tests for existence or not of the property.
    """

    def _fn(value):
        if value is None:
            return None, None
        elif value is True:
            return {"exists": {"field": attr}}, None
        elif value is False:
            return None, {"exists": {"field": attr}}
        elif isinstance(value, list):
            return {"terms": {attr: value}}, None
        else:
            return {"term": {attr: value}}, None

    return _fn


def _gen_filter_name():
    """Match against \"data.name\", or \"case.name\" for case objects."""

    def _fn(value):
        if value is None:
            return None, None
        else:
            return {
                "bool": {
                    "minimum_should_match": 1,
                    "should": [
                        {"term": {"data.name.keyword": value}},
                        {
                            "bool": {
                                "must": [
                                    {"term": {"class.keyword": "case"}},
                                    {"term": {"fmu.case.name.keyword": value}},
                                ]
                            }
                        },
                    ],
                }
            }, None

    return _fn


def _gen_filter_time():
    """Match against a TimeFilter instance."""

    def _fn(value):
        if value is None:
            return None, None
        else:
            return value._get_query(), None

    return _fn


def _gen_filter_bool(attr):
    """Match boolean value."""

    def _fn(value):
        if value is None:
            return None, None
        else:
            return {"term": {attr: value}}, None

    return _fn


def _gen_filter_complex():
    """Match against user-supplied query, which is a structured
    Elasticsearch query in dictionary form."""

    def _fn(value):
        if value is None:
            return None, None
        else:
            return value, None

    return _fn


_filterspec = {
    "id": [_gen_filter_id, None],
    "cls": [_gen_filter_gen, "class.keyword"],
    "time": [_gen_filter_time, None],
    "name": [_gen_filter_name, None],
    "uuid": [_gen_filter_gen, "fmu.case.uuid.keyword"],
    "tagname": [_gen_filter_gen, "data.tagname.keyword"],
    "dataformat": [_gen_filter_gen, "data.format.keyword"],
    "iteration": [_gen_filter_gen, "fmu.iteration.name.keyword"],
    "realization": [_gen_filter_gen, "fmu.realization.id"],
    "aggregation": [_gen_filter_gen, "fmu.aggregation.operation.keyword"],
    "stage": [_gen_filter_gen, "fmu.context.stage.keyword"],
    "column": [_gen_filter_gen, "data.spec.columns.keyword"],
    "vertical_domain": [_gen_filter_gen, "data.vertical_domain.keyword"],
    "content": [_gen_filter_gen, "data.content.keyword"],
    "status": [_gen_filter_gen, "_sumo.status.keyword"],
    "user": [_gen_filter_gen, "fmu.case.user.id.keyword"],
    "asset": [_gen_filter_gen, "access.asset.name.keyword"],
    "field": [_gen_filter_gen, "masterdata.smda.field.identifier.keyword"],
    "stratigraphic": [_gen_filter_bool, "data.stratigraphic"],
    "is_observation": [_gen_filter_bool, "data.is_observation"],
    "is_prediction": [_gen_filter_bool, "data.is_prediction"],
    "complex": [_gen_filter_complex, None],
    "has": [_gen_filter_none, None],
}


def _gen_filters(spec):
    res = {}
    for name, desc in spec.items():
        gen, param = desc
        if param is None:
            res[name] = gen()
        else:
            res[name] = gen(param)
            pass
    return res


filters = _gen_filters(_filterspec)


_bucket_spec = {
    "names": ["data.name.keyword", "List of unique object names."],
    "tagnames": ["data.tagname.keyword", "List of unique object tagnames."],
    "dataformats": [
        "data.format.keyword",
        "List of unique data.format values.",
    ],
    "aggregations": [
        "fmu.aggregation.operation.keyword",
        "List of unique object aggregation operations.",
    ],
    "stages": ["fmu.context.stage.keyword", "List of unique stages."],
    # "stratigraphic": [
    #     "data.stratigraphic",
    #     "List of unique object stratigraphic.",
    # ],
    "vertical_domains": [
        "data.vertical_domain",
        "List of unique object vertical domains.",
    ],
    "contents": ["data.content.keyword", "List of unique contents."],
    "columns": ["data.spec.columns.keyword", "List of unique column names."],
    "statuses": ["_sumo.status.keyword", "List of unique case statuses."],
    "users": ["fmu.case.user.id.keyword", "List of unique user names."],
}


def _build_bucket_query(query, field, size):
    return {
        "size": 0,
        "query": query,
        "aggs": {
            f"{field}": {
                "composite": {
                    "size": size,
                    "sources": [{f"{field}": {"terms": {"field": field}}}],
                }
            }
        },
    }


def _build_bucket_query_simple(query, field, size):
    return {
        "size": 0,
        "query": query,
        "aggs": {f"{field}": {"terms": {"field": field, "size": size}}},
    }


def _set_after_key(query, field, after_key):
    if after_key is not None:
        query["aggs"][field]["composite"]["after"] = after_key
        pass
    return query


def _set_search_after(query, after):
    if after is not None:
        query["search_after"] = after
        pass
    return query


class Pit:
    def __init__(self, sumo: SumoClient, keepalive="5m"):
        self._sumo = sumo
        self._keepalive = keepalive
        self._id = None
        return

    def __enter__(self):
        res = self._sumo.post("/pit", params={"keep-alive": self._keepalive})
        self._id = res.json()["id"]
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._id is not None:
            self._sumo.delete("/pit", params={"id": self._id})
            pass
        return False

    async def __aenter__(self):
        res = await self._sumo.post_async(
            "/pit", params={"keep-alive": self._keepalive}
        )
        self._id = res.json()["id"]
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if self._id is not None:
            await self._sumo.delete_async("/pit", params={"id": self._id})
            pass
        return False

    def stamp_query(self, query):
        query["pit"] = {"id": self._id, "keep_alive": self._keepalive}
        return query

    def update_from_result(self, result):
        self._id = result["pit_id"]
        return


class SearchContext:
    def __init__(
        self,
        sumo: SumoClient,
        must: List = [],
        must_not: List = [],
        hidden=False,
        visible=True,
    ):
        self._sumo = sumo
        self._must = must[:]
        self._must_not = must_not[:]
        self._visible = visible
        self._hidden = hidden
        self._field_values = {}
        self._hits = None
        self._cache = LRUCache(capacity=200)
        self._length = None
        self._select = {
            "excludes": ["fmu.realization.parameters"],
        }
        return

    @property
    def _query(self):
        must = self._must[:]
        must_not = self._must_not[:]
        if self._visible and not self._hidden:
            must_not.append({"term": {"_sumo.hidden": True}})
        elif not self._visible and self._hidden:
            must.append({"term": {"_sumo.hidden": True}})
            pass
        if len(must_not) == 0:
            if len(must) == 1:
                return must[0]
            else:
                return {"bool": {"must": must}}
        else:
            if len(must) == 0:
                return {"bool": {"must_not": must_not}}
            else:
                return {"bool": {"must": must, "must_not": must_not}}

    def _to_sumo(self, obj, blob=None):
        cls = obj["_source"]["class"]
        if cls == "case":
            return objects.Case(self._sumo, obj)
        # ELSE
        constructor = {
            "cube": objects.Cube,
            "dictionary": objects.Dictionary,
            "polygons": objects.Polygons,
            "surface": objects.Surface,
            "table": objects.Table,
            "cpgrid": objects.CPGrid,
            "cpgrid_property": objects.CPGridProperty,
        }.get(cls)
        if constructor is None:
            warnings.warn(f"No constructor for class {cls}")
            constructor = objects.Child
        return constructor(self._sumo, obj, blob)

    def __len__(self):
        if self._hits is not None:
            return len(self._hits)
        if self._length is None:
            query = {"query": self._query}
            res = self._sumo.post("/count", json=query).json()
            self._length = res["count"]
        return self._length

    async def length_async(self):
        if self._hits is not None:
            return len(self._hits)
        if self._length is None:
            query = {"query": self._query}
            res = (await self._sumo.post_async("/count", json=query)).json()
            self._length = res["count"]
        return self._length

    def __search_all(self, query, size=1000, select=False):
        all_hits = []
        query = {
            "query": query,
            "size": size,
            "_source": select,
            "sort": {"_doc": {"order": "asc"}},
            "track_total_hits": True,
        }
        # fast path: try searching without Pit
        res = self._sumo.post("/search", json=query).json()
        total_hits = res["hits"]["total"]["value"]
        if total_hits <= size:
            hits = res["hits"]["hits"]
            if select is False:
                return [hit["_id"] for hit in hits]
            else:
                return hits
        after = None
        with Pit(self._sumo, "1m") as pit:
            while True:
                query = pit.stamp_query(_set_search_after(query, after))
                res = self._sumo.post("/search", json=query).json()
                pit.update_from_result(res)
                hits = res["hits"]["hits"]
                if len(hits) == 0:
                    break
                after = hits[-1]["sort"]
                if select is False:
                    all_hits = all_hits + [hit["_id"] for hit in hits]
                else:
                    all_hits = all_hits + hits
                    pass
                pass
            pass
        return all_hits

    def _search_all(self, select=False):
        return self.__search_all(query=self._query, size=1000, select=select)

    async def __search_all_async(self, query, size=1000, select=False):
        all_hits = []
        query = {
            "query": query,
            "size": size,
            "_source": select,
            "sort": {"_doc": {"order": "asc"}},
            "track_total_hits": True,
        }
        # fast path: try searching without Pit
        res = (await self._sumo.post_async("/search", json=query)).json()
        total_hits = res["hits"]["total"]["value"]
        if total_hits <= size:
            hits = res["hits"]["hits"]
            if select is False:
                return [hit["_id"] for hit in hits]
            else:
                return hits
        after = None
        with Pit(self._sumo, "1m") as pit:
            while True:
                query = pit.stamp_query(_set_search_after(query, after))
                res = (
                    await self._sumo.post_async("/search", json=query)
                ).json()
                pit.update_from_result(res)
                hits = res["hits"]["hits"]
                if len(hits) == 0:
                    break
                after = hits[-1]["sort"]
                if select is False:
                    all_hits = all_hits + [hit["_id"] for hit in hits]
                else:
                    all_hits = all_hits + hits
                    pass
                pass
            pass
        return all_hits

    async def _search_all_async(self, select=False):
        return await self.__search_all_async(
            query=self._query, size=1000, select=select
        )

    @property
    def uuids(self):
        if self._hits is None:
            self._hits = self._search_all()
        return self._hits

    @property
    async def uuids_async(self):
        if self._hits is None:
            self._hits = await self._search_all_async()
        return self._hits

    def __iter__(self):
        self._curr_index = 0
        return self

    def __next__(self):
        if self._hits is None:
            self._hits = self._search_all()
            pass
        if self._curr_index < len(self._hits):
            uuid = self._hits[self._curr_index]
            self._maybe_prefetch(self._curr_index)
            self._curr_index += 1
            return self.get_object(uuid)
        else:
            raise StopIteration

    def __aiter__(self):
        self._curr_index = 0
        return self

    async def __anext__(self):
        if self._hits is None:
            self._hits = await self._search_all_async()
            pass
        if self._curr_index < len(self._hits):
            uuid = self._hits[self._curr_index]
            await self._maybe_prefetch_async(self._curr_index)
            self._curr_index += 1
            return await self.get_object_async(uuid)
        else:
            raise StopAsyncIteration

    def __getitem__(self, index):
        if self._hits is None:
            self._hits = self._search_all()
            pass
        uuid = self._hits[index]
        return self.get_object(uuid)

    async def getitem_async(self, index):
        if self._hits is None:
            self._hits = await self._search_all_async()
            pass
        uuid = self._hits[index]
        return await self.get_object_async(uuid)

    def select(self, sel):
        """Specify what should be returned from elasticsearch.
        Has the side effect of clearing the lru cache.
        sel is either a single string value, a list of string value,
        or a dictionary with keys "includes" and/or "excludes" and
        the values are lists of strings. The string values are nested
        property names.

        Args:
            sel (str | List(str) | Dict(str, List[str]): select specification
        Returns:
            None
        """

        required = {"class"}

        def extreq(lst):
            if isinstance(lst, str):
                lst = [lst]
            return list(set(lst) | required)

        if isinstance(sel, str):
            self._select = extreq([sel])
        elif isinstance(sel, list):
            self._select = extreq(sel)
        elif isinstance(sel, dict):
            inc = sel.get("includes")
            exc = sel.get("excludes")
            slct = {}
            if inc is not None:
                slct["includes"] = extreq(inc)
                pass
            if exc is not None:
                slct["excludes"] = exc
                pass
            self._select = slct
            pass
        self._cache.clear()

    def get_object(self, uuid: str) -> Dict:
        """Get metadata object by uuid

        Args:
            uuid (str): uuid of metadata object
            select (List[str]): list of metadata fields to return

        Returns:
            Dict: a metadata object
        """
        obj = self._cache.get(uuid)
        if obj is None:
            query = {
                "query": {"ids": {"values": [uuid]}},
                "size": 1,
                "_source": self._select,
            }

            res = self._sumo.post("/search", json=query)
            hits = res.json()["hits"]["hits"]

            if len(hits) == 0:
                raise Exception(f"Document not found: {uuid}")
            obj = hits[0]
            self._cache.put(uuid, obj)

        return self._to_sumo(obj)

    async def get_object_async(self, uuid: str) -> Dict:
        """Get metadata object by uuid

        Args:
            uuid (str): uuid of metadata object
            select (List[str]): list of metadata fields to return

        Returns:
            Dict: a metadata object
        """

        obj = self._cache.get(uuid)
        if obj is None:
            query = {
                "query": {"ids": {"values": [uuid]}},
                "size": 1,
                "_source": self._select,
            }

            res = await self._sumo.post_async("/search", json=query)
            hits = res.json()["hits"]["hits"]

            if len(hits) == 0:
                raise Exception(f"Document not found: {uuid}")
            obj = hits[0]
            self._cache.put(uuid, obj)

        return self._to_sumo(obj)

    def _maybe_prefetch(self, index):
        uuid = self._hits[index]
        if self._cache.has(uuid):
            return
        uuids = self._hits[index : min(index + 100, len(self._hits))]
        uuids = [uuid for uuid in uuids if not self._cache.has(uuid)]
        hits = self.__search_all(
            {"ids": {"values": uuids}},
            select=self._select,
        )
        if len(hits) == 0:
            return
        for hit in hits:
            self._cache.put(hit["_id"], hit)
        return

    async def _maybe_prefetch_async(self, index):
        uuid = self._hits[index]
        if self._cache.has(uuid):
            return
        uuids = self._hits[index : min(index + 100, len(self._hits))]
        uuids = [uuid for uuid in uuids if not self._cache.has(uuid)]
        hits = await self.__search_all_async(
            {"ids": {"values": uuids}},
            select=self._select,
        )
        if len(hits) == 0:
            return
        for hit in hits:
            self._cache.put(hit["_id"], hit)
        return

    def get_objects(
        self,
        uuids: List[str],
        select: List[str] = None,
    ) -> List[Dict]:
        size = (
            1000
            if select is False
            else 100
            if isinstance(select, list)
            else 10
        )
        return self.__search_all(
            {"ids": {"values": uuids}}, size=size, select=select
        )

    async def get_objects_async(
        self,
        uuids: List[str],
        select: List[str] = None,
    ) -> List[Dict]:
        size = (
            1000
            if select is False
            else 100
            if isinstance(select, list)
            else 10
        )
        return await self.__search_all_async(
            {"ids": {"values": uuids}}, size=size, select=select
        )

    def _get_buckets(
        self,
        field: str,
    ) -> List[Dict]:
        """Get a List of buckets

        Arguments:
            - field (str): a field in the metadata

        Returns:
            A List of unique values for a given field
        """

        buckets_per_batch = 1000

        # fast path: try without Pit
        query = _build_bucket_query_simple(
            self._query, field, buckets_per_batch
        )
        res = self._sumo.post("/search", json=query).json()
        other_docs_count = res["aggregations"][field]["sum_other_doc_count"]
        if other_docs_count == 0:
            buckets = res["aggregations"][field]["buckets"]
            buckets = [
                {
                    "key": bucket["key"],
                    "doc_count": bucket["doc_count"],
                }
                for bucket in buckets
            ]
            return buckets

        query = _build_bucket_query(self._query, field, buckets_per_batch)
        all_buckets = []
        after_key = None
        with Pit(self._sumo, "1m") as pit:
            while True:
                query = pit.stamp_query(
                    _set_after_key(query, field, after_key)
                )
                res = self._sumo.post("/search", json=query).json()
                pit.update_from_result(res)
                buckets = res["aggregations"][field]["buckets"]
                if len(buckets) == 0:
                    break
                after_key = res["aggregations"][field]["after_key"]
                buckets = [
                    {
                        "key": bucket["key"][field],
                        "doc_count": bucket["doc_count"],
                    }
                    for bucket in buckets
                ]
                all_buckets = all_buckets + buckets
                if len(buckets) < buckets_per_batch:
                    break
                pass

        return all_buckets

    async def _get_buckets_async(
        self,
        field: str,
    ) -> List[Dict]:
        """Get a List of buckets

        Arguments:
            - field (str): a field in the metadata

        Returns:
            A List of unique values for a given field
        """

        buckets_per_batch = 1000

        # fast path: try without Pit
        query = _build_bucket_query_simple(
            self._query, field, buckets_per_batch
        )
        res = (await self._sumo.post_async("/search", json=query)).json()
        other_docs_count = res["aggregations"][field]["sum_other_doc_count"]
        if other_docs_count == 0:
            buckets = res["aggregations"][field]["buckets"]
            buckets = [
                {
                    "key": bucket["key"],
                    "doc_count": bucket["doc_count"],
                }
                for bucket in buckets
            ]
            return buckets

        query = _build_bucket_query(self._query, field, buckets_per_batch)
        all_buckets = []
        after_key = None
        async with Pit(self._sumo, "1m") as pit:
            while True:
                query = pit.stamp_query(
                    _set_after_key(query, field, after_key)
                )
                res = await self._sumo.post_async("/search", json=query)
                res = res.json()
                pit.update_from_result(res)
                buckets = res["aggregations"][field]["buckets"]
                if len(buckets) == 0:
                    break
                after_key = res["aggregations"][field]["after_key"]
                buckets = [
                    {
                        "key": bucket["key"][field],
                        "doc_count": bucket["doc_count"],
                    }
                    for bucket in buckets
                ]
                all_buckets = all_buckets + buckets
                if len(buckets) < buckets_per_batch:
                    break
                pass

        return all_buckets

    def _get_field_values(self, field: str) -> List:
        """Get List of unique values for a given field

        Arguments:
            - field (str): a metadata field

        Returns:
            A List of unique values for the given field
        """
        if field not in self._field_values:
            buckets = self._get_buckets(field)
            self._field_values[field] = [bucket["key"] for bucket in buckets]

        return self._field_values[field]

    async def _get_field_values_async(self, field: str) -> List:
        """Get List of unique values for a given field

        Arguments:
            - field (str): a metadata field

        Returns:
            A List of unique values for the given field
        """
        if field not in self._field_values:
            buckets = await self._get_buckets_async(field)
            self._field_values[field] = [bucket["key"] for bucket in buckets]

        return self._field_values[field]

    _timestamp_query = {
        "bool": {
            "must": [{"exists": {"field": "data.time.t0"}}],
            "must_not": [{"exists": {"field": "data.time.t1"}}],
        }
    }

    def _context_for_class(self, cls):
        return self.filter(cls=cls)

    @property
    def hidden(self):
        return SearchContext(
            sumo=self._sumo,
            must=self._must,
            must_not=self._must_not,
            hidden=True,
            visible=False,
        )

    @property
    def visible(self):
        return SearchContext(
            sumo=self._sumo,
            must=self._must,
            must_not=self._must_not,
            hidden=False,
            visible=True,
        )

    @property
    def all(self):
        return SearchContext(
            sumo=self._sumo,
            must=self._must,
            must_not=self._must_not,
            hidden=True,
            visible=True,
        )

    @property
    def cases(self):
        """Cases from current selection."""
        uuids = self._get_field_values("fmu.case.uuid.keyword")
        return objects.Cases(self, uuids)

    @property
    async def cases_async(self):
        """Cases from current selection."""
        uuids = await self._get_field_values_async("fmu.case.uuid.keyword")
        return objects.Cases(self, uuids)

    @property
    def iterations(self):
        """Iterations from current selection."""
        uuids = self._get_field_values("fmu.iteration.uuid.keyword")
        return objects.Iterations(self, uuids)

    @property
    async def iterations_async(self):
        """Iterations from current selection."""
        uuids = await self._get_field_values_async(
            "fmu.iteration.uuid.keyword"
        )
        return objects.Iterations(self, uuids)

    @property
    def realizations(self):
        """Realizations from current selection."""
        uuids = self._get_field_values("fmu.realization.uuid.keyword")
        return objects.Realizations(self, uuids)

    @property
    async def realizations_async(self):
        """Realizations from current selection."""
        uuids = await self._get_field_values_async(
            "fmu.realization.uuid.keyword"
        )
        return objects.Realizations(self, uuids)

    @property
    def template_paths(search_context):  # noqa: N805
        return {obj.template_path for obj in search_context}

    @property
    def metrics(self):
        """Metrics for current search context."""
        return objects.Metrics(self)

    @property
    def timestamps(self) -> List[str]:
        """List of unique timestamps in SearchContext"""
        ts = self.filter(complex=self._timestamp_query)._get_field_values(
            "data.time.t0.value"
        )
        return [datetime.fromtimestamp(t / 1000).isoformat() for t in ts]

    @property
    async def timestamps_async(self) -> List[str]:
        """List of unique timestamps in SearchContext"""
        ts = await self._get_field_values_async(
            "data.time.t0.value", self._timestamp_query
        )
        return [datetime.fromtimestamp(t / 1000).isoformat() for t in ts]

    def _extract_intervals(self, res):
        buckets = res.json()["aggregations"]["t0"]["buckets"]
        intervals = []

        for bucket in buckets:
            t0 = bucket["key_as_string"]

            for t1 in bucket["t1"]["buckets"]:
                intervals.append((t0, t1["key_as_string"]))

        return intervals

    _intervals_aggs = {
        "t0": {
            "terms": {"field": "data.time.t0.value", "size": 50},
            "aggs": {
                "t1": {
                    "terms": {
                        "field": "data.time.t1.value",
                        "size": 50,
                    }
                }
            },
        }
    }

    @property
    def intervals(self) -> List[Tuple]:
        """List of unique intervals in SearchContext"""
        res = self._sumo.post(
            "/search",
            json={
                "query": self._query,
                "size": 0,
                "aggs": self._intervals_aggs,
            },
        )

        return self._extract_intervals(res)

    @property
    async def intervals_async(self) -> List[Tuple]:
        """List of unique intervals in SearchContext"""
        res = await self._sumo.post_async(
            "/search",
            json={
                "query": self._query,
                "size": 0,
                "aggs": self._intervals_aggs,
            },
        )

        return self._extract_intervals(res)

    def filter(self, **kwargs) -> "SearchContext":
        """Filter SearchContext"""

        must = self._must[:]
        must_not = self._must_not[:]
        for k, v in kwargs.items():
            f = filters.get(k)
            if f is None:
                raise Exception(f"Don't know how to generate filter for {k}")
                pass
            _must, _must_not = f(v)
            if _must:
                must.append(_must)
            if _must_not is not None:
                must_not.append(_must_not)

        sc = SearchContext(
            self._sumo,
            must=must,
            must_not=must_not,
            hidden=self._hidden,
            visible=self._visible,
        )

        if "has" in kwargs:
            # Get list of cases matched by current filter set
            uuids = sc._get_field_values("fmu.case.uuid.keyword")
            # Generate new searchcontext for objects that match the uuids
            # and also satisfy the "has" filter
            sc = SearchContext(
                self._sumo,
                must=[
                    {"terms": {"fmu.case.uuid.keyword": uuids}},
                    kwargs["has"],
                ],
            )
            uuids = sc._get_field_values("fmu.case.uuid.keyword")
            sc = SearchContext(
                self._sumo,
                must=[{"ids": {"values": uuids}}],
            )

        return sc

    @property
    def surfaces(self):
        return self._context_for_class("surface")

    @property
    def tables(self):
        return self._context_for_class("table")

    @property
    def cubes(self):
        return self._context_for_class("cube")

    @property
    def polygons(self):
        return self._context_for_class("polygons")

    @property
    def dictionaries(self):
        return self._context_for_class("dictionary")

    @property
    def grids(self):
        return self._context_for_class("cpgrid")

    @property
    def grid_properties(self):
        return self._context_for_class("cpgrid_property")

    def _get_object_by_class_and_uuid(self, cls, uuid):
        obj = self.get_object(uuid)
        if obj.metadata["class"] != cls:
            raise Exception(f"Document of type {cls} not found: {uuid}")
        return obj

    async def _get_object_by_class_and_uuid_async(self, cls, uuid):
        obj = await self.get_object_async(uuid)
        if obj.metadata["class"] != cls:
            raise Exception(f"Document of type {cls} not found: {uuid}")
        return obj

    def get_case_by_uuid(self, uuid: str):
        """Get case object by uuid

        Args:
            uuid (str): case uuid

        Returns:
            Case: case object
        """
        return self._get_object_by_class_and_uuid("case", uuid)

    async def get_case_by_uuid_async(self, uuid: str):
        """Get case object by uuid

        Args:
            uuid (str): case uuid

        Returns:
            Case: case object
        """
        return await self._get_object_by_class_and_uuid_async("case", uuid)

    def _iteration_query(self, uuid):
        return {
            "query": {"term": {"fmu.iteration.uuid.keyword": {"value": uuid}}},
            "size": 1,
            "_source": {
                "includes": [
                    "$schema",
                    "source",
                    "version",
                    "access",
                    "masterdata",
                    "fmu.case",
                    "fmu.iteration",
                ],
            },
        }

    def get_iteration_by_uuid(self, uuid: str):
        """Get iteration object by uuid

        Args:
            uuid (str): iteration uuid

        Returns: iteration object
        """
        res = self._sumo.post(
            "/search", json=self._iteration_query(uuid)
        ).json()
        obj = res["hits"]["hits"][0]
        obj["_id"] = uuid
        return objects.Iteration(self._sumo, obj)

    async def get_iteration_by_uuid_async(self, uuid: str):
        """Get iteration object by uuid

        Args:
            uuid (str): iteration uuid

        Returns: iteration object
        """
        res = (
            await self._sumo.post_async(
                "/search", json=self._iteration_query(uuid)
            )
        ).json()
        obj = res["hits"]["hits"][0]
        obj["_id"] = uuid
        return objects.Iteration(self._sumo, obj)

    def _realization_query(self, uuid):
        return {
            "query": {
                "term": {"fmu.realization.uuid.keyword": {"value": uuid}}
            },
            "size": 1,
            "_source": {
                "includes": [
                    "$schema",
                    "source",
                    "version",
                    "access",
                    "masterdata",
                    "fmu.case",
                    "fmu.iteration",
                    "fmu.realization",
                ],
            },
        }

    def get_realization_by_uuid(self, uuid: str):
        """Get realization object by uuid

        Args:
            uuid (str): realization uuid

        Returns: realization object
        """
        res = self._sumo.post(
            "/search", json=self._realization_query(uuid)
        ).json()
        obj = res["hits"]["hits"][0]
        obj["_id"] = uuid
        return objects.Realization(self._sumo, obj)

    async def get_realization_by_uuid_async(self, uuid: str):
        """Get realization object by uuid

        Args:
            uuid (str): realization uuid

        Returns: realization object
        """
        res = (
            await self._sumo.post(
                "/search", json=self._realization_query(uuid)
            )
        ).json()
        obj = res["hits"]["hits"][0]
        obj["_id"] = uuid
        return objects.Realization(self._sumo, obj)

    def get_surface_by_uuid(self, uuid: str):
        """Get surface object by uuid

        Args:
            uuid (str): surface uuid

        Returns:
            Surface: surface object
        """
        return self._get_object_by_class_and_uuid("surface", uuid)

    async def get_surface_by_uuid_async(self, uuid: str):
        """Get surface object by uuid

        Args:
            uuid (str): surface uuid

        Returns:
            Surface: surface object
        """
        return await self._get_object_by_class_and_uuid_async("surface", uuid)

    def get_polygons_by_uuid(self, uuid: str):
        """Get polygons object by uuid

        Args:
            uuid (str): polygons uuid

        Returns:
            Polygons: polygons object
        """
        return self._get_object_by_class_and_uuid("polygons", uuid)

    async def get_polygons_by_uuid_async(self, uuid: str):
        """Get polygons object by uuid

        Args:
            uuid (str): polygons uuid

        Returns:
            Polygons: polygons object
        """
        return await self._get_object_by_class_and_uuid_async("polygons", uuid)

    def get_table_by_uuid(self, uuid: str):
        """Get table object by uuid

        Args:
            uuid (str): table uuid

        Returns:
            Table: table object
        """
        return self._get_object_by_class_and_uuid("table", uuid)

    async def get_table_by_uuid_async(self, uuid: str):
        """Get table object by uuid

        Args:
            uuid (str): table uuid

        Returns:
            Table: table object
        """
        return await self._get_object_by_class_and_uuid_async("table", uuid)

    def _verify_aggregation_operation(self):
        query = {
            "query": self._query,
            "size": 1,
            "track_total_hits": True,
            "aggs": {
                k: {"terms": {"field": k + ".keyword", "size": 1}}
                for k in [
                    "fmu.case.uuid",
                    "class",
                    "fmu.iteration.name",
                    "data.name",
                    "data.tagname",
                    "data.content",
                ]
            },
        }
        sres = self._sumo.post("/search", json=query).json()
        prototype = sres["hits"]["hits"][0]
        conflicts = [
            k
            for (k, v) in sres["aggregations"].items()
            if (
                ("sum_other_doc_count" in v) and (v["sum_other_doc_count"] > 0)
            )
        ]
        if len(conflicts) > 0:
            raise Exception(f"Conflicting values for {conflicts}")

        hits = self._search_all(select=["fmu.realization.id"])

        if any(
            hit["_source"]["fmu"].get("realization") is None for hit in hits
        ):
            raise Exception("Selection contains non-realization data.")

        uuids = [hit["_id"] for hit in hits]
        rids = [hit["_source"]["fmu"]["realization"]["id"] for hit in hits]
        return prototype, uuids, rids

    def _aggregate(self, columns=None, operation=None):
        prototype, uuids, rids = self._verify_aggregation_operation()
        spec = {
            "object_ids": uuids,
            "operations": [operation],
        }
        del prototype["_source"]["fmu"]["realization"]
        del prototype["_source"]["_sumo"]
        del prototype["_source"]["file"]
        del prototype["_source"]["access"]
        if "context" in prototype["_source"]["fmu"]:
            prototype["_source"]["fmu"]["context"]["stage"] = "iteration"
            pass
        prototype["_source"]["fmu"]["aggregation"] = {
            "id": str(uuid.uuid4()),
            "realization_ids": rids,
            "operation": operation,
        }
        if columns is not None:
            spec["columns"] = columns
            cols = columns[:]
            table_index = prototype["_source"]["data"].get("table_index")
            if (
                table_index is not None
                and len(table_index) != 0
                and table_index[0] not in cols
            ):
                cols.insert(0, table_index[0])
                pass
            prototype["_source"]["data"]["spec"]["columns"] = cols
            pass
        try:
            res = self._sumo.post("/aggregations", json=spec)
        except httpx.HTTPStatusError as ex:
            print(ex.response.reason_phrase)
            print(ex.response.text)
            raise ex
        blob = BytesIO(res.content)
        res = self._to_sumo(prototype, blob)
        res._blob = blob
        return res

    def aggregate(self, columns=None, operation=None):
        if len(self.hidden) > 0:
            return self.hidden._aggregate(columns=columns, operation=operation)
        else:
            return self.visible._aggregate(
                columns=columns, operation=operation
            )

    async def _verify_aggregation_operation_async(self):
        query = {
            "query": self._query,
            "size": 1,
            "track_total_hits": True,
            "aggs": {
                k: {"terms": {"field": k + ".keyword", "size": 1}}
                for k in [
                    "fmu.case.uuid",
                    "class",
                    "fmu.iteration.name",
                    "data.name",
                    "data.tagname",
                    "data.content",
                ]
            },
        }
        sres = (await self._sumo.post_async("/search", json=query)).json()
        prototype = sres["hits"]["hits"][0]
        conflicts = [
            k
            for (k, v) in sres["aggregations"].items()
            if (
                ("sum_other_doc_count" in v) and (v["sum_other_doc_count"] > 0)
            )
        ]
        if len(conflicts) > 0:
            raise Exception(f"Conflicting values for {conflicts}")

        hits = await self._search_all_async(select=["fmu.realization.id"])

        if any(
            hit["_source"]["fmu"].get("realization") is None for hit in hits
        ):
            raise Exception("Selection contains non-realization data.")

        uuids = [hit["_id"] for hit in hits]
        rids = [hit["_source"]["fmu"]["realization"]["id"] for hit in hits]
        return prototype, uuids, rids

    async def _aggregate_async(self, columns=None, operation=None):
        (
            prototype,
            uuids,
            rids,
        ) = await self._verify_aggregation_operation_async()
        spec = {
            "object_ids": uuids,
            "operations": [operation],
        }
        del prototype["_source"]["fmu"]["realization"]
        del prototype["_source"]["_sumo"]
        del prototype["_source"]["file"]
        del prototype["_source"]["access"]
        if "context" in prototype["_source"]["fmu"]:
            prototype["_source"]["fmu"]["context"]["stage"] = "iteration"
            pass
        prototype["_source"]["fmu"]["aggregation"] = {
            "id": str(uuid.uuid4()),
            "realization_ids": rids,
            "operation": operation,
        }
        if columns is not None:
            spec["columns"] = columns
            cols = columns[:]
            table_index = prototype["_source"]["data"].get("table_index")
            if (
                table_index is not None
                and len(table_index) != 0
                and table_index[0] not in cols
            ):
                cols.insert(0, table_index[0])
                pass
            prototype["_source"]["data"]["spec"]["columns"] = cols
            pass
        try:
            res = await self._sumo.post_async("/aggregations", json=spec)
        except httpx.HTTPStatusError as ex:
            print(ex.response.reason_phrase)
            print(ex.response.text)
            raise ex
        blob = BytesIO(res.content)
        res = self._to_sumo(prototype, blob)
        res._blob = blob
        return res

    async def aggregate_async(self, columns=None, operation=None):
        length_hidden = await self.hidden.length_async()
        if length_hidden > 0:
            return await self.hidden._aggregate_async(
                columns=columns, operation=operation
            )
        else:
            return await self.visible._aggregate_async(
                columns=columns, operation=operation
            )

    def aggregation(self, column=None, operation=None):
        assert operation is not None
        assert column is None or isinstance(column, str)
        sc = self.filter(aggregation=operation, column=column)
        numaggs = len(sc)
        assert numaggs <= 1
        if numaggs == 1:
            return sc[0]
        else:
            return self.filter(realization=True).aggregate(
                columns=[column] if column is not None else None,
                operation=operation,
            )

    async def aggregation_async(self, column=None, operation=None):
        assert operation is not None
        assert column is None or isinstance(column, str)
        sc = self.filter(aggregation=operation, column=column)
        numaggs = await sc.length_async()
        assert numaggs <= 1
        if numaggs == 1:
            return await sc.getitem_async(0)
        else:
            return await self.filter(realization=True).aggregate_async(
                columns=[column] if column is not None else None,
                operation=operation,
            )

    @deprecation.deprecated(
        details="Use the method 'aggregate' instead, with parameter 'operation'."
    )
    def min(self):
        return self.aggregate(operation="min")

    @deprecation.deprecated(
        details="Use the method 'aggregate' instead, with parameter 'operation'."
    )
    def max(self):
        return self.aggregate(operation="max")

    @deprecation.deprecated(
        details="Use the method 'aggregate' instead, with parameter 'operation'."
    )
    def mean(self):
        return self.aggregate(operation="mean")

    @deprecation.deprecated(
        details="Use the method 'aggregate' instead, with parameter 'operation'."
    )
    def std(self):
        return self.aggregate(operation="std")

    @deprecation.deprecated(
        details="Use the method 'aggregate' instead, with parameter 'operation'."
    )
    def p10(self):
        return self.aggregate(operation="p10")

    @deprecation.deprecated(
        details="Use the method 'aggregate' instead, with parameter 'operation'."
    )
    def p50(self):
        return self.aggregate(operation="p50")

    @deprecation.deprecated(
        details="Use the method 'aggregate' instead, with parameter 'operation'."
    )
    def p90(self):
        return self.aggregate(operation="p90")


def _gen_filter_doc(spec):
    fmap = {
        _gen_filter_id: "Id",
        _gen_filter_bool: "Boolean",
        _gen_filter_name: "Name",
        _gen_filter_gen: "General",
        _gen_filter_time: "Time",
        _gen_filter_complex: "Complex",
    }
    ret = """\
Filter SearchContext.

Apply additional filters to SearchContext and return a new
filtered instance.

The filters (specified as keyword args) are of these formats:

"""
    for gen, name in fmap.items():
        ret = ret + f"    {name}:  {gen.__doc__}\n"
    ret = (
        ret
        + """
Args:

"""
    )
    for name in sorted(spec.keys()):
        gen, property = spec[name]
        if gen in [_gen_filter_complex, _gen_filter_none]:
            continue
        typ = fmap.get(gen)
        if typ is not None:
            if property is None:
                ret = ret + f"    {name} ({typ})\n"
            else:
                ret = ret + f'    {name} ({typ}): "{property}"\n'
                pass
            pass
        pass
    ret = ret + "    has (Complex)\n"
    ret = ret + "    complex (Complex)\n"
    ret = (
        ret
        + """
Returns:
    SearchContext: A filtered SearchContext.

Examples:

    Match one value::

        surfs = case.surfaces.filter(
                    iteration="iter-0",
                    name="my_surface_name"
                )

    Match multiple values::

        surfs = case.surfaces.filter(
                    name=["one_name", "another_name"]
                )

    Get aggregated surfaces with specific operation::

        surfs = case.surfaces.filter(
                    aggregation="max"
                )

    Get all aggregated surfaces::

        surfs = case.surfaces.filter(
                    aggregation=True
                )

    Get all non-aggregated surfaces::

        surfs = case.surfaces.filter(
                    aggregation=False
                )

"""
    )
    return ret


SearchContext.filter.__doc__ = _gen_filter_doc(_filterspec)


def _build_bucket_fn(property, docstring):
    def fn(self):
        return self._get_field_values(property)

    return fn


def _build_bucket_fn_async(property, docstring):
    async def fn(self):
        return await self._get_field_values_async(property)

    return fn


def _inject_bucket_fns(spec):
    for name, defn in spec.items():
        prop, docstring = defn
        fn = _build_bucket_fn(prop, docstring)
        setattr(SearchContext, name, property(fn, None, None, docstring))
        afn = _build_bucket_fn_async(prop, docstring)
        setattr(
            SearchContext,
            name + "_async",
            property(afn, None, None, docstring),
        )
        pass
    return


_inject_bucket_fns(_bucket_spec)



"""module containing class for child object"""

from io import BytesIO
from typing import Dict

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._document import Document

_prop_desc = [
    ("name", "data.name", "Object name"),
    ("dataname", "data.name", "Object name"),
    ("classname", "class.name", "Object class name"),
    ("casename", "fmu.case.name", "Object case name"),
    ("content", "data.content", "Content"),
    ("tagname", "data.tagname", "Object tagname"),
    ("columns", "data.spec.columns", "Object table columns"),
    ("stratigraphic", "data.stratigraphic", "Object stratigraphic"),
    ("vertical_domain", "data.vertical_domain", "Object vertical domain"),
    ("context", "fmu.context.stage", "Object context"),
    ("iteration", "fmu.iteration.name", "Object iteration"),
    ("realization", "fmu.realization.id", "Object realization"),
    (
        "aggregation",
        "fmu.aggregation.operation",
        "Object aggregation operation",
    ),
    ("stage", "fmu.context.stage", "Object stage"),
    ("format", "data.format", "Object file format"),
    ("dataformat", "data.format", "Object file format"),
    ("relative_path", "file.relative_path", "Object relative file path"),
    ("bbox", "data.bbox", "Object boundary-box data"),
    ("spec", "data.spec", "Object spec data"),
]


class Child(Document):
    """Class representing a child object in Sumo"""

    def __init__(self, sumo: SumoClient, metadata: Dict, blob=None) -> None:
        """
        Args:
            sumo (SumoClient): connection to Sumo
            metadata: (dict): child object metadata
        """
        super().__init__(metadata)
        self._sumo = sumo
        self._blob = blob

    @property
    def blob(self) -> BytesIO:
        """Object blob"""
        if self._blob is None:
            res = self._sumo.get(f"/objects('{self.uuid}')/blob")
            self._blob = BytesIO(res.content)

        return self._blob

    @property
    async def blob_async(self) -> BytesIO:
        """Object blob"""
        if self._blob is None:
            res = await self._sumo.get_async(f"/objects('{self.uuid}')/blob")
            self._blob = BytesIO(res.content)

        return self._blob

    @property
    def timestamp(self) -> str:
        """Object timestmap data"""
        t0 = self._get_property(["data", "time", "t0", "value"])
        t1 = self._get_property(["data", "time", "t1", "value"])

        if t0 is not None and t1 is None:
            return t0

        return None

    @property
    def interval(self) -> str:
        """Object interval data"""
        t0 = self._get_property(["data", "time", "t0", "value"])
        t1 = self._get_property(["data", "time", "t1", "value"])

        if t0 is not None and t1 is not None:
            return (t0, t1)

        return None

    @property
    def template_path(self):
        return "/".join(
            ["{realization}", "{iteration}"]
            + self.relative_path.split("/")[2:]
        )


Child.map_properties(Child, _prop_desc)



"""Module containing class for cpgrid_property"""

from typing import Dict

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._child import Child
from fmu.sumo.explorer.objects._search_context import SearchContext


class CPGridProperty(Child):
    """Class representing a cpgrid_property object in Sumo."""

    def __init__(self, sumo: SumoClient, metadata: Dict, blob=None) -> None:
        """
        Args:
            sumo (SumoClient): connection to Sumo
            metadata (dict): dictionary metadata
            blob: data object
        """
        super().__init__(sumo, metadata, blob)

    def to_cpgrid_property(self):
        """Get cpgrid_property object as a GridProperty
        Returns:
            GridProperty: A GridProperty object
        """
        try:
            from xtgeo import gridproperty_from_file
        except ModuleNotFoundError:
            raise RuntimeError(
                "Unable to import xtgeo; probably not installed."
            )
        try:
            return gridproperty_from_file(self.blob)
        except TypeError as type_err:
            raise TypeError(f"Unknown format: {self.format}") from type_err

    async def to_cpgrid_property_async(self):
        """Get cpgrid_property object as a GridProperty
        Returns:
            GridProperty: A GridProperty object
        """
        try:
            from xtgeo import gridproperty_from_file
        except ModuleNotFoundError:
            raise RuntimeError(
                "Unable to import xtgeo; probably not installed."
            )

        try:
            return gridproperty_from_file(await self.blob_async)
        except TypeError as type_err:
            raise TypeError(f"Unknown format: {self.format}") from type_err

    @property
    def grid(self):
        """Get cpgrid object associated with this cpgrid_property instances.
        Returns:
            Grid: a Grid object (an instance of class CPGrid).
        """
        sc = SearchContext(self._sumo).filter(
            complex={
                "bool": {
                    "must": [
                        {
                            "term": {
                                "file.relative_path.keyword": self._metadata[
                                    "data"
                                ]["geometry"]["relative_path"]
                            }
                        },
                        {
                            "term": {
                                "fmu.case.uuid.keyword": self._metadata["fmu"][
                                    "case"
                                ]["uuid"]
                            }
                        },
                    ]
                }
            }
        )
        assert len(sc) == 1
        return sc[0]



class Metrics:
    def __init__(self, search_context):
        self._search_context = search_context
        return

    def _aggregate(self, op, **kwargs):
        aggs = {
            "agg": {op: {k: v for k, v in kwargs.items() if v is not None}}
        }
        qdoc = {"query": self._search_context._query, "aggs": aggs, "size": 0}
        res = self._search_context._sumo.post("/search", json=qdoc).json()
        return res["aggregations"]["agg"]

    def min(self, field):
        """Find the minimum value for the specified property across the
        current set of objects.

        Arguments:
            - field (str): the name of a property in the metadata.

        Returns:
            The minimum value.

        """
        return self._aggregate("min", field=field)["value"]

    def max(self, field):
        """Find the maximum value for the specified property across the
        current set of objects.

        Arguments:
            - field (str): the name of a property in the metadata.

        Returns:
            The maximum value.

        """
        return self._aggregate("max", field=field)["value"]

    def avg(self, field):
        """Find the average value for the specified property across the
        current set of objects.

        Arguments:
            - field (str): the name of a property in the metadata.

        Returns:
            The average value.

        """
        return self._aggregate("avg", field=field)["value"]

    def sum(self, field):
        """Find the sumo of all values for the specified property across
        the current set of objects.

        Arguments:
            - field (str): the name of a property in the metadata.

        Returns:
            The sum of all values.

        """
        return self._aggregate("sum", field=field)["value"]

    def value_count(self, field):
        """Find the count of values for the specified property across the
        current set of objects.

        Arguments:
            - field (str): the name of a property in the metadata.

        Returns:
            The total number of values.

        """
        return self._aggregate("value_count", field=field)["value"]

    def cardinality(self, field):
        """Find the count of distinct values for the specified property
        across the current set of objects.

        Arguments:
            - field (str): the name of a property in the metadata.

        Returns:
            The number of distinct values.

        Note: The value returned is approximate.

        """
        return self._aggregate("cardinality", field=field)["value"]

    def stats(self, field):
        """Compute a basic set of statistics of the values for the specified
        property across the current set of objects.

        Arguments:
            - field (str): the name of a property in the metadata.

        Returns:
            A dictionary of statistical metrics.

        """
        return self._aggregate("stats", field=field)

    def extended_stats(self, field):
        """Compute an extended set of statistics of the values for the
        specified property across the current set of objects.

        Arguments:
            - field (str): the name of a property in the metadata.

        Returns:
            A dictionary of statistical metrics.

        """
        return self._aggregate("extended_stats", field=field)

    def percentiles(self, field, percents=None):
        """Find the values at specific percentiles for the specified
        property across the current set of objects.

        Arguments:
            - field (str): the name of a property in the metadata.
            - percents ([number]): list of percent values. If omitted, uses
              a default set of values.

        Returns:
            A dictionary of percentiles.

        """
        return self._aggregate("percentiles", field=field, percents=percents)[
            "values"
        ]



"""Module for (pseudo) iteration class."""

from typing import Dict

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._document import Document
from fmu.sumo.explorer.objects._search_context import SearchContext

_prop_desc = [
    ("name", "fmu.iteration.name", "FMU iteration name"),
    ("casename", "fmu.case.name", "FMU case name"),
    ("caseuuid", "fmu.case.uuid", "FMU case uuid"),
    ("user", "fmu.case.user.id", "Name of user who uploaded iteration."),
    ("asset", "access.asset.name", "Case asset"),
    ("field", "masterdata.smda.field[0].identifier", "Case field"),
]


class Iteration(Document, SearchContext):
    """Class for representing an iteration in Sumo."""

    def __init__(self, sumo: SumoClient, metadata: Dict):
        Document.__init__(self, metadata)
        SearchContext.__init__(
            self,
            sumo,
            must=[{"term": {"fmu.iteration.uuid.keyword": self.uuid}}],
        )


Iteration.map_properties(Iteration, _prop_desc)



"""Sumo cases and child objects"""

from fmu.sumo.explorer.objects._child import Child
from fmu.sumo.explorer.objects._metrics import Metrics
from fmu.sumo.explorer.objects._search_context import SearchContext
from fmu.sumo.explorer.objects.case import Case
from fmu.sumo.explorer.objects.cases import Cases
from fmu.sumo.explorer.objects.cpgrid import CPGrid
from fmu.sumo.explorer.objects.cpgrid_property import CPGridProperty
from fmu.sumo.explorer.objects.cube import Cube
from fmu.sumo.explorer.objects.dictionary import Dictionary
from fmu.sumo.explorer.objects.iteration import Iteration
from fmu.sumo.explorer.objects.iterations import Iterations
from fmu.sumo.explorer.objects.polygons import Polygons
from fmu.sumo.explorer.objects.realization import Realization
from fmu.sumo.explorer.objects.realizations import Realizations
from fmu.sumo.explorer.objects.surface import Surface
from fmu.sumo.explorer.objects.table import Table



"""Module containing class for dictionary object"""

import json
from typing import Dict

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._child import Child


class Dictionary(Child):
    """Class representig a dictionary object in Sumo"""

    _parsed: dict

    def __init__(self, sumo: SumoClient, metadata: Dict, blob=None) -> None:
        """
        Args:
            sumo (SumoClient): connection to Sumo
            metadata (dict): dictionary metadata
        """
        self._parsed = None

        super().__init__(sumo, metadata, blob)

    def parse(self) -> Dict:
        if self._parsed is None:
            self._parsed = json.loads(self.blob.decode("utf-8"))

        return self._parsed



"""Module containg class for surface"""

from typing import Dict

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._child import Child


class Surface(Child):
    """Class representing a surface object in Sumo"""

    def __init__(self, sumo: SumoClient, metadata: Dict, blob=None) -> None:
        """
        Args:
            sumo (SumoClient): connection to Sumo
            metadata (dict): dictionary metadata
            blob: data object
        """
        super().__init__(sumo, metadata, blob)

    def to_regular_surface(self):
        """Get surface object as a RegularSurface

        Returns:
            RegularSurface: A RegularSurface object
        """
        try:
            from xtgeo import surface_from_file
        except ModuleNotFoundError:
            raise RuntimeError(
                "Unable to import xtgeo; probably not installed."
            )

        try:
            return surface_from_file(self.blob)
        except TypeError as type_err:
            raise TypeError(f"Unknown format: {self.format}") from type_err

    async def to_regular_surface_async(self):
        """Get surface object as a RegularSurface

        Returns:
            RegularSurface: A RegularSurface object
        """
        try:
            from xtgeo import surface_from_file
        except ModuleNotFoundError:
            raise RuntimeError(
                "Unable to import xtgeo; probably not installed."
            )

        try:
            return surface_from_file(await self.blob_async)
        except TypeError as type_err:
            raise TypeError(f"Unknown format: {self.format}") from type_err



"""Test access to SUMO using a DROGON-MANAGE login.
Shall only run in Github Actions as a specific user with
specific access rights. Running this test with your personal login
will fail."""

import inspect
import json
import os

import pytest
from context import (
    Explorer,
)

if os.getenv("GITHUB_ACTIONS") == "true":
    RUNNING_OUTSIDE_GITHUB_ACTIONS = "False"
    print(
        "Found the GITHUB_ACTIONS env var, so I know I am running on Github now. Will run these tests."
    )
else:
    RUNNING_OUTSIDE_GITHUB_ACTIONS = "True"
    msg = "Skipping these tests since they can only run on Github Actions as a specific user"
    print("NOT running on Github now.", msg)
    pytest.skip(msg, allow_module_level=True)


@pytest.fixture(name="explorer")
def fixture_explorer(token: str) -> Explorer:
    """Returns explorer"""
    return Explorer("dev", token=token)


def test_admin_access(explorer: Explorer):
    """Test access to an admin endpoint"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    with pytest.raises(Exception, match="403*"):
        print("About to call an admin endpoint which should raise exception")
        explorer._sumo.get(
            "/admin/make-shared-access-key?user=noreply%40equinor.com&roles=DROGON-READ&duration=111"
        )
        print("Execution should never reach this line")


def test_get_userpermissions(explorer: Explorer):
    """Test the userpermissions"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    response = explorer._sumo.get("/userpermissions")
    print("/Userpermissions response: ", response.text)
    userperms = json.loads(response.text)
    assert "Drogon" in userperms
    assert "manage" in userperms.get("Drogon")
    assert len(userperms.get("Drogon")) == 1
    assert len(userperms) == 1


def test_get_cases(explorer: Explorer):
    """Test the get_cases method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    cases = explorer.cases
    print("Number of cases: ", len(cases))
    for case in cases:
        assert case.field.lower() == "drogon"
    assert len(cases) > 0


def test_write(explorer: Explorer):
    """Test a write method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    cases = explorer.cases.filter(status="scratch")
    print("Number of cases: ", len(cases))
    assert len(cases) > 0
    case = cases[0]
    print("case uuid:", case.metadata.get("fmu").get("case").get("uuid"))
    print("About to write to a case")
    response = explorer._sumo.post("/objects", json=case.metadata)
    print(response.status_code)
    print(response.text)
    assert response.status_code == 200


def test_read_restricted_classification_data(explorer: Explorer):
    """Test if can read restriced data aka 'access:classification: restricted'"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    cases = explorer.cases
    print("Number of cases: ", len(cases))
    assert len(cases) > 0

    # A default Drogon iteration contains 2 restricted objects,
    # so in normal situations there should be some restricted objects
    response = explorer._sumo.get(
        "/search?%24query=access.classification%3Arestricted"
    )
    assert response.status_code == 200
    response_json = json.loads(response.text)
    hits = response_json.get("hits").get("total").get("value")
    print("Hits on restricted:", hits)
    assert hits > 0


# Remove or update this test when bulk aggregation is finalized
# @pytest.mark.skipif(not (sys.platform == "linux" and
#                          sys.version_info[:2] == (3, 11)),
#                     reason="Test only on single platform/version.")
# def test_aggregations_bulk(explorer: Explorer):
#     """Test a bulk aggregation method"""
#     print("Running test:", inspect.currentframe().f_code.co_name)
#     cases = explorer.cases.filter(status="scratch")
#     print("Number of cases: ", len(cases))
#     assert len(cases) > 0
#     case = None
#     for c in cases:
#         if len(c.realizations) > 1 and len(c.surfaces) > 40:
#             case = c
#             break
#     assert case
#     case_uuid = case.metadata.get("fmu").get("case").get("uuid")
#     print("About to trigger aggregation on case", case_uuid)
#     body = {
#         "operations": ["min"],
#         "case_uuid": case_uuid,
#         "class": "surface",
#         "iteration_name": case.iterations[0].name,
#     }
#     response = explorer._sumo.post(f"/aggregations", json=body)
#     print(response.status_code)
#     assert response.status_code in [200, 201, 202]


def test_aggregations_fast(explorer: Explorer):
    """Test a fast aggregation method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    cases = explorer.cases
    print("Number of cases: ", len(cases))
    assert len(cases) > 0
    case = None
    for c in cases:
        if (
            len(c.realizations) > 1
            and len(c.surfaces) > 40
            and len(c.iterations) == 1
            and len(
                c.surfaces.filter(
                    name="Therys Fm.", tagname="FACIES_Fraction_Calcite"
                )
            )
            > 2
        ):
            case = c
            break
    assert case
    case_uuid = case.metadata.get("fmu").get("case").get("uuid")
    print("About to trigger fast-aggregation on case", case_uuid)
    surface1 = case.surfaces.filter(
        name="Therys Fm.", realization=0, tagname="FACIES_Fraction_Calcite"
    )
    surface2 = case.surfaces.filter(
        name="Therys Fm.", realization=1, tagname="FACIES_Fraction_Calcite"
    )
    print("Len filtered: ", len(surface1))
    print("Len filtered: ", len(surface2))
    assert len(surface1) == 1
    assert len(surface2) == 1
    surface_uuids = [surface1[0].uuid, surface2[0].uuid]
    body = {
        "operations": ["min"],
        "object_ids": surface_uuids,
        "class": "surface",
        "iteration_name": case.iterations[0].name,
    }
    response = explorer._sumo.post("/aggregations", json=body)
    print("Response status code:", response.status_code)
    assert response.status_code == 200
    print("Length of returned aggregate object:", len(response.text))


def test_get_access_log(explorer: Explorer):
    """Test to get the access log method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    print("About to get access log")
    response = explorer._sumo.get("/access-log")
    print(response.status_code)
    print(len(response.text))
    # Currently all authenticated users have access
    assert response.status_code == 200


def test_get_key(explorer: Explorer):
    """Test to get key method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    print("About to get key, which should raise exception ")
    with pytest.raises(Exception, match="403*"):
        response = explorer._sumo.get("/key")
        print("Execution should never reach this line")
        print("Unexpected status: ", response.status_code)
        print("Unexpected response: ", response.text)


def test_get_purge(explorer: Explorer):
    """Test to get purge method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    print("About to get purge, which should raise exception ")
    with pytest.raises(Exception, match="403*"):
        response = explorer._sumo.get("/purge")
        print("Execution should never reach this line")
        print("Unexpected status: ", response.status_code)
        print("Unexpected response: ", response.text)


def test_get_message_log_truncate(explorer: Explorer):
    """Test to get msg log truncate method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    print("About to get msg log truncate, which should raise exception ")
    with pytest.raises(Exception, match="403*"):
        response = explorer._sumo.get("/message-log/truncate?cutoff=99")
        print("Execution should never reach this line")
        print("Unexpected status: ", response.status_code)
        print("Unexpected response: ", response.text)



"""Module containing case class"""

from typing import Dict

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._document import Document
from fmu.sumo.explorer.objects._search_context import SearchContext

_prop_desc = [
    ("name", "fmu.case.name", "Case name"),
    ("status", "_sumo.status", "Case status"),
    ("user", "fmu.case.user.id", "Name of user who uploaded case."),
    ("asset", "access.asset.name", "Case asset"),
    ("field", "masterdata.smda.field[0].identifier", "Case field"),
]


def _make_overview_query(id):
    return {
        "query": {"term": {"fmu.case.uuid.keyword": id}},
        "aggs": {
            "iteration_uuids": {
                "terms": {"field": "fmu.iteration.uuid.keyword", "size": 100}
            },
            "iteration_names": {
                "terms": {"field": "fmu.iteration.name.keyword", "size": 100}
            },
            "data_types": {"terms": {"field": "class.keyword", "size": 100}},
            "iterations": {
                "terms": {"field": "fmu.iteration.uuid.keyword", "size": 100},
                "aggs": {
                    "iteration_name": {
                        "terms": {
                            "field": "fmu.iteration.name.keyword",
                            "size": 100,
                        }
                    },
                    "numreal": {
                        "cardinality": {"field": "fmu.realization.id"}
                    },
                    "maxreal": {"max": {"field": "fmu.realization.id"}},
                    "minreal": {"min": {"field": "fmu.realization.id"}},
                },
            },
        },
        "size": 0,
    }


class Case(Document, SearchContext):
    """Class for representing a case in Sumo"""

    def __init__(self, sumo: SumoClient, metadata: Dict):
        Document.__init__(self, metadata)
        SearchContext.__init__(
            self, sumo, must=[{"term": {"fmu.case.uuid.keyword": self.uuid}}]
        )
        self._overview = None
        self._iterations = None

    @property
    def overview(self):
        """Overview of case contents."""

        def extract_bucket_keys(bucket, name):
            return [b["key"] for b in bucket[name]["buckets"]]

        if self._overview is None:
            query = _make_overview_query(self._uuid)
            res = self._sumo.post("/search", json=query)
            data = res.json()
            aggs = data["aggregations"]
            iteration_names = extract_bucket_keys(aggs, "iteration_names")
            iteration_uuids = extract_bucket_keys(aggs, "iteration_uuids")
            data_types = extract_bucket_keys(aggs, "data_types")
            iterations = {}
            for bucket in aggs["iterations"]["buckets"]:
                iterid = bucket["key"]
                itername = extract_bucket_keys(bucket, "iteration_name")
                minreal = bucket["minreal"]["value"]
                maxreal = bucket["maxreal"]["value"]
                numreal = bucket["numreal"]["value"]
                iterations[iterid] = {
                    "name": itername,
                    "minreal": minreal,
                    "maxreal": maxreal,
                    "numreal": numreal,
                }
            self._overview = {
                "iteration_names": iteration_names,
                "iteration_uuids": iteration_uuids,
                "data_types": data_types,
                "iterations": iterations,
            }

        return self._overview


Case.map_properties(Case, _prop_desc)



"""Test access to SUMO using a DROGON-AFFILIATE login.
Shall only run in Github Actions as a specific user with
specific access rights. Running this test with your personal login
will fail."""

import inspect
import json
import os

import pytest
from context import (
    Explorer,
)

if os.getenv("GITHUB_ACTIONS") == "true":
    RUNNING_OUTSIDE_GITHUB_ACTIONS = "False"
    print(
        "Found the GITHUB_ACTIONS env var, so I know I am running on Github now. Will run these tests."
    )
else:
    RUNNING_OUTSIDE_GITHUB_ACTIONS = "True"
    msg = "Skipping these tests since they can only run on Github Actions as a specific user"
    print("NOT running on Github now.", msg)
    pytest.skip(msg, allow_module_level=True)


@pytest.fixture(name="explorer")
def fixture_explorer(token: str) -> Explorer:
    """Returns explorer"""
    return Explorer("dev", token=token)


def test_admin_access(explorer: Explorer):
    """Test access to an admin endpoint"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    with pytest.raises(Exception, match="403*"):
        print("About to call an admin endpoint which should raise exception")
        explorer._sumo.get(
            "/admin/make-shared-access-key?user=noreply%40equinor.com&roles=DROGON-READ&duration=111"
        )
        print("Execution should never reach this line")


def test_get_userpermissions(explorer: Explorer):
    """Test the userpermissions"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    response = explorer._sumo.get("/userpermissions")
    print("/Userpermissions response: ", response.text)
    userperms = json.loads(response.text)
    assert "Drogon" in userperms
    assert "affiliate" in userperms.get("Drogon")
    assert len(userperms.get("Drogon")) == 1
    assert len(userperms) == 1


def test_get_cases(explorer: Explorer):
    """Test the get_cases method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    cases = explorer.cases
    print("Number of cases: ", len(cases))
    for case in cases:
        assert case.field.lower() == "drogon"
    assert len(cases) >= 1

    # We have set up 1 case in KEEP in Drogon DEV
    # with affiliate-access and it has 2 children
    # objects with affiliate access
    filtered_cases = cases.filter(uuid="2c2f47cf-c7ab-4112-87f9-b4797ec51cb6")
    assert len(filtered_cases) == 1
    case = filtered_cases[0]
    assert case.uuid == "2c2f47cf-c7ab-4112-87f9-b4797ec51cb6"
    assert len(case.surfaces) == 1
    assert len(case.polygons) == 1
    assert len(case.tables) == 0
    assert len(case.cubes) == 0
    assert len(case.dictionaries) == 0
    assert case.polygons[0].uuid == "a5f38286-5cf6-d85c-9b3c-03c72b5947d5"
    assert case.surfaces[0].uuid == "5f73b0c1-3bdc-2d0e-1a1d-271331615999"

    # Many Drogon cases might have been shared now, who knows.
    # Ensure that all returned cases have correct metadata:
    for case in cases:
        affiliate_roles = case._metadata.get("access").get("affiliate_roles")
        assert "DROGON-AFFILIATE" in affiliate_roles


def test_get_object(explorer: Explorer):
    """Test the direct get on object by objectid method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    cases = explorer.cases
    print("Number of cases: ", len(cases))

    # We have set up a KEEP case in Drogon DEV with
    # objects with affiliate-access

    # Read one child object
    child_object_uuid = "a5f38286-5cf6-d85c-9b3c-03c72b5947d5"
    response = explorer._sumo.get(f"/objects('{child_object_uuid}')")
    print("child retval:", response)
    print("child retval.content:", response.content)
    assert response.status_code == 200
    response_json = json.loads(response.text)
    child_uuid = response_json.get("_id")
    print("child_uuid returned:", child_uuid)
    assert child_uuid == child_object_uuid
    classification = (
        response_json.get("_source").get("access").get("classification")
    )
    assert classification == "internal"

    # Read the other child object (which also have
    # access.classification:restricted)
    child_object_uuid = "5f73b0c1-3bdc-2d0e-1a1d-271331615999"
    response = explorer._sumo.get(f"/objects('{child_object_uuid}')")
    print("child retval:", response)
    print("child retval.content:", response.content)
    assert response.status_code == 200
    response_json = json.loads(response.text)
    child_uuid = response_json.get("_id")
    print("child_uuid returned:", child_uuid)
    assert child_uuid == child_object_uuid
    classification = (
        response_json.get("_source").get("access").get("classification")
    )
    assert classification == "restricted"

    # Read the case object
    case_object_uuid = "2c2f47cf-c7ab-4112-87f9-b4797ec51cb6"
    response = explorer._sumo.get(f"/objects('{case_object_uuid}')")
    print("case retval:", response)
    print("case retval.content:", response.content)
    assert response.status_code == 200
    response_json = json.loads(response.text)
    case_uuid = response_json.get("_id")
    print("case_uuid returned:", case_uuid)
    assert case_uuid == case_object_uuid

    # Read an object which is NOT for affiliates
    child_wo_access = "6d9d222f-25d3-5029-07b8-f450dbd2be54"
    with pytest.raises(Exception, match="404*"):
        print("About to call a endpoint which should raise exception")
        response = explorer._sumo.get(f"/objects('{child_wo_access}')")
        print("Execution should never reach this line")
        print("Unexpected status: ", response.status_code)
        print("Unexpected response: ", response.text)

    # The exact number of shared Drogon cases cannot be known
    assert len(cases) >= 1


def test_delete(explorer: Explorer):
    """Test a delete method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    cases = explorer.cases
    print("Number of cases: ", len(cases))

    for case in cases:
        if case.uuid == "392c3c70-dd1a-41b5-ac49-0e369a0ac4eb":
            print("Found my case")
            surface_uuid = case.surfaces[0].uuid
            print("Surface:", surface_uuid)

            res = explorer._sumo.get(f"/objects('{surface_uuid}')")
            print("Get Result should be OK:", res.json())

            with pytest.raises(Exception, match="403*"):
                res = explorer._sumo.delete(f"/objects('{surface_uuid}')")
                print("Execution should never reach this line")
                print("Unexpected status: ", res.status_code)
                print("Unexpected response: ", res.text)

            with pytest.raises(Exception, match="403*"):
                res = explorer._sumo.delete(f"/objects('{case.uuid}')")
                print("Execution should never reach this line")
                print("Unexpected status: ", res.status_code)
                print("Unexpected response: ", res.text)


def test_write(explorer: Explorer):
    """Test a write method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    cases = explorer.cases
    print("Number of cases: ", len(cases))

    with open("./tests/data/test_case_080/case2.json") as json_file:
        metadata = json.load(json_file)
        with pytest.raises(Exception, match="403*"):
            print(
                "About to call a write endpoint which should raise exception"
            )
            response = explorer._sumo.post("/objects", json=metadata)
            print("Execution should never reach this line")
            print("Unexpected status: ", response.status_code)
            print("Unexpected response: ", response.text)


def test_read_restricted_classification_data(explorer: Explorer):
    """Test if can read restricted data aka 'access:classification: restricted'"""
    print("Running test:", inspect.currentframe().f_code.co_name)

    # access.classification:restricted is available,
    # EVEN for this DROGON-AFFILIATE user
    # (This differs from DROGON-READ which cannot read restricted)
    response = explorer._sumo.get(
        "/search?%24query=access.classification%3Arestricted"
    )
    assert response.status_code == 200
    response_json = json.loads(response.text)
    hits = response_json.get("hits").get("total").get("value")
    print("Hits on restricted:", hits)
    assert hits >= 1


# Remove or update this test when bulk aggregation is finalized
# @pytest.mark.skipif(not (sys.platform == "linux" and
#                          sys.version_info[:2] == (3, 11)),
#                     reason="Test only on single platform/version.")
# def test_aggregate_bulk(explorer: Explorer):
#     """Test a bulk aggregation method"""
#     print("Running test:", inspect.currentframe().f_code.co_name)
#     # Fixed test case ("Drogon_AHM_2023-02-22") in Sumo/DEV
#     # This user does not have WRITE and hence this should fail
#     # (In addition this case is not set up with affiliate access)
#     TESTCASE_UUID = "10f41041-2c17-4374-a735-bb0de62e29dc"
#     print("About to trigger bulk aggregation on case", TESTCASE_UUID)
#     body = {
#         "operations": ["min"],
#         "case_uuid": TESTCASE_UUID,
#         "class": "surface",
#         "iteration_name": "iter-0",
#     }
#     with pytest.raises(Exception, match="40*"):
#         response = explorer._sumo.post(f"/aggregations", json=body)
#         print("Execution should never reach this line")
#         print("Unexpected status: ", response.status_code)
#         print("Unexpected response: ", response.text)


def test_aggregations_fast(explorer: Explorer):
    """Test a fast aggregation method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    # Fixed test case ("Drogon_AHM_2023-02-22") in Sumo/DEV
    # This user has AFFILIATE and can READ, but this case
    # is not set up with AFFILIATE access, so should fail
    testcase_uuid = "10f41041-2c17-4374-a735-bb0de62e29dc"
    print("About to trigger fast-aggregation on case", testcase_uuid)
    surface_uuid_1 = "ae6cf480-12ba-77ca-848e-92e707556b63"
    surface_uuid_2 = "7189835b-cc8a-2a8e-4a34-dde2ceb2a69c"
    body = {
        "operations": ["min"],
        "object_ids": [surface_uuid_1, surface_uuid_2],
        "class": "surface",
        "iteration_name": "iter-0",
    }
    print("About to trigger fast-aggregation on hardcoded case", testcase_uuid)
    print("using body", body)
    with pytest.raises(Exception, match="40*"):
        response = explorer._sumo.post("/aggregations", json=body)
        print("Execution should never reach this line")
        print("Unexpected status: ", response.status_code)
        print("Unexpected response: ", response.text)


# TODO: TBC: Consider setting up a case with affiliate access on all
# surfaces, so we can test successful fast aggregation. Need first
# a clarification if affiliates are allowed fast aggregation or not


def test_get_access_log(explorer: Explorer):
    """Test to get the access log method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    print("About to get access log")
    response = explorer._sumo.get("/access-log")
    print(response.status_code)
    print(len(response.text))
    # Currently all authenticated users have access
    assert response.status_code == 200


def test_get_key(explorer: Explorer):
    """Test to get key method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    print("About to get key, which should raise exception ")
    with pytest.raises(Exception, match="403*"):
        response = explorer._sumo.get("/key")
        print("Execution should never reach this line")
        print("Unexpected status: ", response.status_code)
        print("Unexpected response: ", response.text)


def test_get_purge(explorer: Explorer):
    """Test to get purge method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    print("About to get purge, which should raise exception ")
    with pytest.raises(Exception, match="403*"):
        response = explorer._sumo.get("/purge")
        print("Execution should never reach this line")
        print("Unexpected status: ", response.status_code)
        print("Unexpected response: ", response.text)


def test_get_message_log_truncate(explorer: Explorer):
    """Test to get msg log truncate method"""
    print("Running test:", inspect.currentframe().f_code.co_name)
    print("About to get msg log truncate, which should raise exception ")
    with pytest.raises(Exception, match="403*"):
        response = explorer._sumo.get("/message-log/truncate?cutoff=99")
        print("Execution should never reach this line")
        print("Unexpected status: ", response.status_code)
        print("Unexpected response: ", response.text)
